# Created by Scott David Daniels on Wed, 23 Jun 2004, licensed under the PSF
# http://code.activestate.com/recipes/286165-ilines-universal-newlines-from-any-data-source/

def ilines(source_iterable):
    '''yield lines as in universal-newlines from a stream of data blocks'''
    tail = ''
    for block in source_iterable:
        if not block:
            continue
        if tail.endswith('\015'):
            yield tail[:-1] + '\012'
            if block.startswith('\012'):
                pos = 1
            else:
                tail = ''
        else:
            pos = 0
        try:
            while True: # While we are finding LF.
                npos = block.index('\012', pos) + 1
                try:
                    rend = npos - 2
                    rpos = block.index('\015', pos, rend)
                    if pos:
                        yield block[pos : rpos] + '\n'
                    else:
                        yield tail + block[:rpos] + '\n'
                    pos = rpos + 1
                    while True: # While CRs 'inside' the LF
                        rpos = block.index('\015', pos, rend)
                        yield block[pos : rpos] + '\n'
                        pos = rpos + 1
                except ValueError:
                    pass
                if '\015' == block[rend]:
                    if pos:
                        yield block[pos : rend] + '\n'
                    else:
                        yield tail + block[:rend] + '\n'
                elif pos:
                    yield block[pos : npos]
                else:
                    yield tail + block[:npos]
                pos = npos
        except ValueError:
            pass
        # No LFs left in block.  Do all but final CR (in case LF)
        try:
            while True:
                rpos = block.index('\015', pos, -1)
                if pos:
                    yield block[pos : rpos] + '\n'
                else:
                    yield tail + block[:rpos] + '\n'
                pos = rpos + 1
        except ValueError:
            pass

        if pos:
            tail = block[pos:]
        else:
            tail += block
    if tail:
        yield tail



try:
    # python 2.7:
    from collections import OrderedDict
except ImportError:
    ## {{{ http://code.activestate.com/recipes/576669/ (r18)
    ## Raymond Hettingers proporsal to go in 2.7
    from collections import MutableMapping

    class OrderedDict(dict, MutableMapping):

        # Methods with direct access to underlying attributes

        def __init__(self, *args, **kwds):
            if len(args) > 1:
                raise TypeError('expected at 1 argument, got %d', len(args))
            if not hasattr(self, '_keys'):
                self._keys = []
            self.update(*args, **kwds)

        def clear(self):
            del self._keys[:]
            dict.clear(self)

        def __setitem__(self, key, value):
            if key not in self:
                self._keys.append(key)
            dict.__setitem__(self, key, value)

        def __delitem__(self, key):
            dict.__delitem__(self, key)
            self._keys.remove(key)

        def __iter__(self):
            return iter(self._keys)

        def __reversed__(self):
            return reversed(self._keys)

        def popitem(self):
            if not self:
                raise KeyError
            key = self._keys.pop()
            value = dict.pop(self, key)
            return key, value

        def __reduce__(self):
            items = [[k, self[k]] for k in self]
            inst_dict = vars(self).copy()
            inst_dict.pop('_keys', None)
            return (self.__class__, (items,), inst_dict)

        # Methods with indirect access via the above methods

        setdefault = MutableMapping.setdefault
        update = MutableMapping.update
        pop = MutableMapping.pop
        keys = MutableMapping.keys
        values = MutableMapping.values
        items = MutableMapping.items

        def __repr__(self):
            pairs = ', '.join(map('%r: %r'.__mod__, self.items()))
            return '%s({%s})' % (self.__class__.__name__, pairs)

        def copy(self):
            return self.__class__(self)

        @classmethod
        def fromkeys(cls, iterable, value=None):
            d = cls()
            for key in iterable:
                d[key] = value
            return d
    ## end of http://code.activestate.com/recipes/576669/ }}}

def offset_processor(offset):
    """ Skip ``offset`` from the given iterator. This can 
    be used in combination with the ``headers_processor`` to 
    apply the result of a header scan to the table. """
    i = [0]
    def apply_offset(row_set, row):
        if i[0] >= offset:
            return row
        i[0] += 1
    return apply_offset



import decimal
from datetime import datetime
from collections import defaultdict
#from pprint import pprint

from messytables.dateparser import DATE_FORMATS

class CellType(object):
    """ A cell type maintains information about the format 
    of the cell, providing methods to check if a type is 
    applicable to a given value and to convert a value to the
    type. """

    guessing_weight = 1

    @classmethod
    def test(cls, value):
        """ Test if the value is of the given type. The 
        default implementation calls ``cast`` and checks if 
        that throws an exception. """
        try:
            ins = cls()
            ins.cast(value)
            return ins
        except: 
            return None

    def cast(self, value):
        """ Convert the value to the type. This may throw
        a quasi-random exception if conversion fails. """
        return value

    def __eq__(self, other):
        return self.__class__ == other.__class__

    def __hash__(self):
        return hash(self.__class__)

    def __repr__(self):
        return self.__class__.__name__.rsplit('Type', 1)[0]

class StringType(CellType):
    """ A string or other unconverted type. """

    def cast(self, value):
        if not isinstance(value, basestring):
            raise ValueError()
        return value

class IntegerType(CellType):
    """ An integer field. """
    guessing_weight = 1.5

    def cast(self, value):
        return int(value)

class FloatType(CellType):
    """ Floating point number. """
    guessing_weight = 2

    def cast(self, value):
        return float(value)

class DecimalType(CellType):
    """ Decimal number, ``decimal.Decimal``. """
    guessing_weight = 2.5

    def cast(self, value):
        return decimal.Decimal(value)

class DateType(CellType):
    """ The date type is special in that it also includes a specific
    date format that is used to parse the date, additionally to the
    basic type information. """
    guessing_weight = 3
    formats = DATE_FORMATS

    def __init__(self, format):
        self.format = format

    @classmethod
    def test(cls, value):
        for v in cls.formats:
            ins = cls(v)
            try:
                ins.cast(value)
                return ins
            except: pass

    def cast(self, value):
        if self.format is None:
            return value
        return datetime.strptime(value, self.format)

    def __eq__(self, other):
        return isinstance(other, DateType) and \
                self.format == other.format
    
    def __repr__(self):
        return "Date(%s)" % self.format

    def __hash__(self):
        return hash(self.__class__) + hash(self.format)

TYPES = [StringType, IntegerType, FloatType, DecimalType, DateType]

def type_guess(rows, types=TYPES):
    """ The type guesser aggregates the number of successful 
    conversions of each column to each type, weights them by a 
    fixed type priority and select the most probable type for 
    each column based on that figure. It returns a list of 
    ``CellType``. """
    guesses = defaultdict(lambda: defaultdict(int))
    for row in rows:
        for i, cell in enumerate(row):
            for type in types:
                guess = type.test(cell.value)
                if guess is not None:
                    guesses[i][guess] += 1
    for i, guess in guesses.items():
        for type, count in guess.items():
            guesses[i][type] *= type.guessing_weight
    _columns = []
    for i, types in guesses.items():
        _columns.append(max(types.items(), key=lambda (t, n): n)[0])
    return _columns

from itertools import izip_longest
def types_processor(types):
    """ Apply the column types set on the instance to the
    current row, attempting to cast each cell to the specified
    type. """
    def apply_types(row_set, row):
        if types is None:
            return row
        for cell, type in izip_longest(row, types):
            try:
                cell.value = type.cast(cell.value)
                cell.type = type
            except:
                pass
        return row
    return apply_types



from collections import defaultdict

from messytables.core import Cell

def column_count_modal(rows):
    """ Return the modal value of columns in the row_set's 
    sample. This can be assumed to be the number of columns
    of the table. """
    counts = defaultdict(int)
    for row in rows:
        length = len([c for c in row if not c.empty])
        counts[length] += 1
    if not len(counts):
        return 0
    return max(counts.items(), key=lambda (k,v): v)[0]

def headers_guess(rows, tolerance=1):
    """ Guess the offset and names of the headers of the row set.
    This will attempt to locate the first row within ``tolerance``
    of the mode of the number of rows in the row set sample.

    The return value is a tuple of the offset of the header row
    and the names of the columns.
    """
    rows = list(rows)
    modal = column_count_modal(rows)
    for i, row in enumerate(rows):
        length = len([c for c in row if not c.empty])
        if length >= modal - tolerance:
            # TODO: use type guessing to check that this row has
            # strings and does not conform to the type schema of 
            # the table.
            return i,  [c.value for c in row]
    return 0, []

from itertools import izip_longest
def headers_processor(headers):
    """ Add column names to the cells in a row_set. If no header is 
    defined, use an autogenerated name. """
    def apply_headers(row_set, row):
        _row = []
        pairs = izip_longest(row, headers)
        for i, (cell, header) in enumerate(pairs):
            if cell is None:
                cell = Cell(None)
            cell.column = header
            if cell.column is None or not len(cell.column):
                cell.column = "column_%d" % i
                cell.column_autogenerated = True
            _row.append(cell)
        return _row
    return apply_headers




from tempfile import mkstemp
from datetime import datetime
from shutil import copyfileobj
from itertools import islice
import xlrd

from messytables.core import RowSet, TableSet, Cell
from messytables.types import StringType, IntegerType, \
        DateType

XLS_TYPES = {
    # TODO: extend with float etc.
    1: StringType(), 
    2: IntegerType(),
    3: DateType(None)
    }

class XLSTableSet(TableSet):
    """ An excel workbook wrapper object. As the underlying
    library is based on reading from a file name (as opposed to
    a file object), a local, temporary copy is created and 
    passed into the library. This has significant performance 
    implication for large excel sheets. """

    def __init__(self, filename):
        self.workbook = xlrd.open_workbook(filename)

    @classmethod
    def from_fileobj(cls, fileobj):
        """ Create a local copy of the object and attempt 
        to open it with xlrd. """
        fd, name = mkstemp(suffix='xls')
        copyfileobj(fileobj, open(name, 'wb'))
        return cls(name)

    @property
    def tables(self):
        """ Return the sheets in the workbook. """
        return [XLSRowSet(name, self.workbook.sheet_by_name(name)) \
                for name in self.workbook.sheet_names()]


class XLSRowSet(RowSet):
    """ Excel support for a single sheet in the excel workbook. Unlike
    the CSV row set this is not a streaming operation. """

    def __init__(self, name, sheet, window=1000):
        self.name = name
        self.sheet = sheet
        self.window = window
        super(XLSRowSet, self).__init__(typed=True)

    @property
    def sample(self):
        return islice(self, self.window)

    def raw(self, sample=False):
        """ Iterate over all rows in this sheet. Types are automatically
        converted according to the excel data types specified, including 
        conversion of excel dates, which are notoriously buggy. """
        num_rows = self.sheet.nrows
        for i in xrange(min(self.window, num_rows) if sample else num_rows):
            row = []
            for cell in self.sheet.row(i):
                value = cell.value
                type = XLS_TYPES.get(cell.ctype, StringType())
                if type == DateType(None):
                    year, month, day, hour, minute, second = \
                            xlrd.xldate_as_tuple(value, self.sheet.book.datemode)
                    value = datetime(year, month, day, hour, 
                            minute, second)
                row.append(Cell(value, type=type))
            yield row

